@@ -13,7 +13,6 @@ from mypy.nodes import (
     UNBOUND_IMPORTED,
     Expression,
     MypyFile,
-    Node,
     SymbolTable,
     SymbolTableNode,
     TypeInfo,
@@ -103,7 +102,6 @@ class ASTMergeSuite(DataSuite):
         options.export_types = True
         options.show_traceback = True
         options.allow_empty_bodies = True
-        options.force_uppercase_builtins = True
         main_path = os.path.join(test_temp_dir, "main")
 
         self.str_conv.options = options
@@ -172,10 +170,7 @@ class ASTMergeSuite(DataSuite):
             if node.kind == UNBOUND_IMPORTED:
                 return "UNBOUND_IMPORTED"
             return "None"
-        if isinstance(node.node, Node):
-            s = f"{str(type(node.node).__name__)}<{self.id_mapper.id(node.node)}>"
-        else:
-            s = f"? ({type(node.node)})"
+        s = f"{str(type(node.node).__name__)}<{self.id_mapper.id(node.node)}>"
         if (
             isinstance(node.node, Var)
             and node.node.type
